import {
  ConduitRoute,
  ConduitRouteReturnDefinition,
  PlatformTypesEnum,
} from '@conduitplatform/commons';
import {
  ConduitError,
  ConduitRouteActions,
  ConduitRouteParameters,
  ConduitString,
} from '@conduitplatform/grpc-sdk';
import { Client } from '../../models';
import { isNil } from 'lodash';

export function getUpdateSecurityClientRoute() {
  const { clientSecret, ...returnTypeFields } = Client.getInstance().fields;
  return new ConduitRoute(
    {
      path: '/security/client/:id',
      urlParams: {
        id: ConduitString.Required,
      },
      action: ConduitRouteActions.UPDATE,
      bodyParams: {
        domain: ConduitString.Optional,
        alias: ConduitString.Optional,
        notes: ConduitString.Optional,
      },
    },
    new ConduitRouteReturnDefinition('UpdateSecurityClient', returnTypeFields),
    async (params: ConduitRouteParameters) => {
      const { domain, alias, notes } = params.params!;
      let client = await Client.getInstance().findOne({
        _id: params.params!.id,
      });
      if (isNil(client)) {
        throw new ConduitError('INVALID_PARAMS', 400, 'Security client not found');
      }
      if (alias === '') {
        throw new ConduitError(
          'INVALID_ARGUMENTS',
          400,
          'Non-null alias field should not be an empty string',
        );
      }
      if (alias && alias !== client.alias) {
        const existingClient = await Client.getInstance().findOne({ alias });
        if (existingClient) {
          throw new ConduitError(
            'ALREADY_EXISTS',
            409,
            `A security client with an alias of '${alias}' already exists`,
          );
        }
      }
      if (client.platform === PlatformTypesEnum.WEB) {
        if (!domain || domain === '')
          throw new ConduitError(
            'INVALID_ARGUMENTS',
            400,
            'Platform WEB requires domain name',
          );
        if (domain.replace(/[^*]/g, '').length > 1) {
          throw new ConduitError(
            'INVALID_ARGUMENTS',
            400,
            `Domain must not contain more than one '*' character`,
          );
        }
        const domainPattern = new RegExp(
          '^(?!-)[A-Za-z0-9-]+([\\-\\.]{1}[a-z0-9]+)*\\.[A-Za-z]{2,6}$',
        );
        let comparedDomain = domain;
        if (domain.includes('*')) {
          comparedDomain = comparedDomain.split('*.')[1];
        }
        if (!domainPattern.test(comparedDomain) && domain !== '*')
          throw new ConduitError('INVALID_ARGUMENTS', 400, 'Invalid domain argument');
      }
      client = await Client.getInstance().findByIdAndUpdate(client._id, {
        domain: client.platform === PlatformTypesEnum.WEB ? domain : undefined,
        alias,
        notes,
      });
      return {
        result: client,
      }; // unnested from result in Rest.addConduitRoute, grpc routes avoid this using wrapRouterGrpcFunction
    },
  );
}
